#include "impls/cyclic/laplacexz-cyclic.hxx"
#include "impls/petsc/laplacexz-petsc.hxx"

#include <bout/invert/laplacexz.hxx>

// DO NOT REMOVE: ensures linker keeps all symbols in this TU
void LaplaceXZFactory::ensureRegistered() {}
