#ifndef BOUT_SOLVER_FACTORY_H
#define BOUT_SOLVER_FACTORY_H

#ifndef _MSC_VER
#warning("Deprecated header: use #include <bout/solver.hxx> instead")
#else
#pragma message("Warning: deprecated header: use #include <bout/solver.hxx> instead")
#endif

#include <bout/solver.hxx>

#endif // BOUT_SOLVER_FACTORY_H
